import cgi
from http import client as http_client
from typing import Any, Dict

import requests

from . import base, http


Json = Dict[str, Any]

DEFAULT_JSON_CONTENT_TYPE = "application/json"


class JSONError(base.ClientError):
    """Exception encountered retrieving JSON content."""

    pass


class JSONClient(http.HTTPClient[Json]):
    """Base class for JSON clients."""

    # Default JSON response back
    default_content_type = DEFAULT_JSON_CONTENT_TYPE

    def _deserialize(self, response: requests.Response) -> Json:
        return _deserialize_json_response(response, valid_content_type=self.default_content_type)


# Helpers:


def _deserialize_json_response(
    response: requests.Response, valid_content_type: str = DEFAULT_JSON_CONTENT_TYPE
) -> Json:
    content_type, _ = cgi.parse_header(response.headers.get("content-type", ""))

    if content_type == valid_content_type:
        return response.json()

    # When 204 is returned, there is explicitly no content.
    if response.status_code == http_client.NO_CONTENT:
        return {}

    raise JSONError(f"No JSON content returned: {response.text}")



import abc
from typing import Dict, Generic, Optional, TypeVar


T = TypeVar("T")

Header = Dict[str, str]

DEFAULT_TIMEOUT = 10.0


class ClientError(Exception):
    """Basic exception raised by an HTTP client."""

    pass


class BaseClient(abc.ABC, Generic[T]):
    """Interface class for HTTP clients."""

    _url: str
    _timeout: float
    _headers: Header

    # Default base URL
    default_base_url: str = ""

    # Default connection timeout
    default_timeout: float = DEFAULT_TIMEOUT

    # Default request media type
    default_content_type: str = ""

    def __init__(
        self,
        base_url: Optional[str] = None,
        default_timeout: Optional[float] = None,
        default_headers: Optional[Header] = None,
    ) -> None:
        self._url = base_url or self.default_base_url
        self._timeout = default_timeout or self.default_timeout
        self._headers = default_headers or {"Accept": self.default_content_type}

    # Default HTTP methods:

    def post(self, path: str, data: Optional[dict] = None, statuses: tuple = ()) -> T:
        return self._request(method="POST", path=path, data=data, statuses=statuses)

    def put(self, path: str, data: Optional[dict] = None, statuses: tuple = ()) -> T:
        return self._request(method="PUT", path=path, data=data, statuses=statuses)

    def patch(self, path: str, data: Optional[dict] = None, statuses: tuple = ()) -> T:
        return self._request(method="PATCH", path=path, data=data, statuses=statuses)

    def get(self, path: str, params: Optional[dict] = None, statuses: tuple = ()) -> T:
        return self._request(method="GET", path=path, params=params, statuses=statuses)

    def delete(self, path: str, params: Optional[dict] = None, statuses: tuple = ()) -> T:
        return self._request(method="DELETE", path=path, params=params, statuses=statuses)

    # Bespoke client implementation:

    @abc.abstractmethod
    def _request(
        self,
        method: str,
        path: str,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        statuses: tuple = (),
    ) -> T:
        ...



from typing import Optional, Type

from requests import PreparedRequest, auth
from typing_extensions import Protocol


DEFAULT_TOKEN_HEADER = "X-TokenAuth"
DEFAULT_TOKEN_VALUE = "Token"


class HTTPTokenAuth(auth.AuthBase):
    """Attaches a bearer token authentication header."""

    # Default token header and value
    default_header = DEFAULT_TOKEN_HEADER
    default_value = DEFAULT_TOKEN_VALUE

    def __init__(
        self, token: str, header: Optional[str] = None, value: Optional[str] = None
    ) -> None:
        self._token = token
        self._header = header or self.default_header
        self._value = value or self.default_value

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        request.headers[self._header] = f"{self._value} {self._token}"
        return request


class AuthClient(Protocol):
    """Mixin for authenticated clients (for token- or basic-based auth)."""

    _auth: Optional[auth.AuthBase] = None

    # Default token authentication class
    default_auth_class: Type[auth.AuthBase] = auth.HTTPBasicAuth

    def set_auth_header(self, *args) -> None:
        """Set authentication header."""
        self._auth = self.default_auth_class(*args)



import abc
from typing import Optional, TypeVar
from urllib import parse

import requests
from requests import auth

from . import base


T = TypeVar("T")


class HTTPError(base.ClientError):
    """Exception encountered over a HTTP client connection."""

    pass


class HTTPClient(base.BaseClient[T]):
    """Base class for blocking HTTP clients."""

    _auth: Optional[auth.AuthBase] = None
    _conn: Optional[requests.Session] = None

    @property
    def conn(self) -> requests.Session:
        """HTTP client connection object."""
        if self._conn is None:
            self._conn = requests.Session()
            self._conn.headers.update(self._headers)
        return self._conn

    # Blocking client implementation:

    def _request(
        self,
        method: str,
        path: str,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        statuses: tuple = (),
    ) -> T:
        url = parse.urljoin(self._url, path)
        response = self._send(method, url, params=params, data=data)

        self._check(response, statuses=statuses)
        return self._deserialize(response)

    # Helpers:

    def _send(
        self, method: str, url: str, params: Optional[dict] = None, data: Optional[dict] = None
    ) -> requests.Response:
        """Fetch request response."""
        return self.conn.request(
            method,
            url,
            params=params,
            json=data,
            headers=self._headers,
            timeout=self._timeout,
            auth=self._auth,
        )

    @staticmethod
    def _check(response: requests.Response, statuses: tuple = ()) -> None:
        """Check response status."""
        _check_response_status(response, statuses=statuses)

    @abc.abstractmethod
    def _deserialize(self, response: requests.Response) -> T:
        """Deserialize response content."""
        ...


# Helpers:


def _check_response_status(response: requests.Response, statuses: tuple = ()) -> None:
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError:
        raise HTTPError(f"{response.status_code}: {response.text}")
    else:
        if response.status_code not in statuses:
            codes = ",".join([str(status) for status in statuses])
            raise HTTPError(f"Unexpected {response.status_code} response: expected {codes}")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['noos_pyk', 'noos_pyk.clients']

package_data = \
{'': ['*']}

install_requires = \
['requests']

setup_kwargs = {
    'name': 'noos-pyk',
    'version': '0.0.2',
    'description': 'A simple, yet useful Python toolkit',
    'long_description': '[![CircleCI](https://circleci.com/gh/noosenergy/noos-python-kit.svg?style=svg&circle-token=5c5370df196704b1e8a8dd7c6f2ec0731c154beb)](https://circleci.com/gh/noosenergy/noos-python-kit)\n\n# Noos Energy Python Toolkit\n\nThis is a simple, yet useful toolkit that supports you in writing microservices-style Python apps.\n\n## Installation\n\nPackage available from the [PyPi repository](https://pypi.org/project/noos-pyk/):\n\n    $ pip install noos-pyk\n\n## Usage as a library\n\nThe project currently houses a boilerplate to build Python clients for REST services.\n\nAs an example, to implement a Python client wrapping up HashiCorp\'s Terraform Cloud API,\n\n```python\n# Import the namespace within your project\nfrom noos_pyk.clients import auth, json\n\n\n# Define a bearer token authentication class\nclass TerraformAuth(auth.HTTPTokenAuth):\n    default_header = "Authorization"\n    default_value = "Bearer"\n\n\n# Wireup all components for a JSON REST client\nclass TerraformClient(json.JSONClient, auth.AuthClient):\n    default_base_url = "https://app.terraform.io/api/"\n    default_content_type = "application/vnd.api+json"\n\n    default_auth_class = TerraformAuth\n```\n\n## Development\n\nOn Mac OSX, make sure [poetry](https://python-poetry.org/) has been installed and pre-configured,\n\n    $ brew install poetry\n\nThis project is shipped with a Makefile, which is ready to do basic common tasks.\n\n```\n$ make\n\nhelp                           Display this auto-generated help message\nclean                          Clean project from temp files / dirs\nformat                         Run auto-formatting linters\nlint                           Run python linters\ntest                           Run pytest with all tests\npackage                        Build project wheel distribution\nrelease                        Publish wheel distribution to PyPi\n```\n',
    'author': 'Noos Energy',
    'author_email': 'contact@noos.energy',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/noosenergy/noos-python-kit',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
